import traceback
import click
import json
import logging
import sys

from click.core import ParameterSource
import paho.mqtt.publish as mqtt
from textwrap import dedent

from .protocol import Protocol
from .usbutils import get_device_serial, reset_device, find_device_by_serial


def resolve_serial_port(serial: str) -> str:
    """Resolve the serial parameter to a device path.

    If the parameter looks like a device path (contains / or starts with COM),
    return it as-is. Otherwise, treat it as a USB serial number and find the
    corresponding device.

    Args:
        serial: Either a device path or USB serial number

    Returns:
        The device path to use
    """
    # Check if it looks like a device path
    if '/' in serial or serial.upper().startswith('COM'):
        return serial

    # Otherwise, treat it as a USB serial number
    return find_device_by_serial(serial)


@click.group()
def cli():
    pass


@cli.command()
@click.option('--serial', '-s', required=True, help='Serial port (e.g., COM1, /dev/ttyUSB0) or USB device serial number')
def test(serial):
    """Try to read one status update"""
    serial_port = resolve_serial_port(serial)
    protocol = Protocol(serial_port)
    for battery_status in protocol.read_update():
        click.echo(repr(battery_status.to_dict()))


@cli.command()
@click.option('--serial', '-s', required=True, help='Serial port (e.g., COM1, /dev/ttyUSB0) or USB device serial number')
@click.option('--hostname', '-h', required=True, help="MQTT hostname")
@click.option('--topic_prefix', '-t', help="MQTT topic prefix", default="pylon2mqtt")
@click.option('--publish_absent', is_flag=True, help="Publish absent battery slots", default=False)
@click.option('--first_only', is_flag=True, default=False,
              help="Only publish the first battery (directly to --topic_prefix)")
@click.option('--update_interval', '-i', help="Update interval in seconds", type=float, default=60.0)
def run(serial: str, hostname: str, topic_prefix: str, publish_absent: bool, first_only: bool, update_interval: float):
    """Read values from a Pylon battery and publish to MQTT periodically"""
    serial_port = resolve_serial_port(serial)
    protocol = Protocol(serial_port)
    for battery_states in protocol.read_updates(interval_secs=update_interval):
        messages = []
        for battery_state in battery_states:
            if battery_state.is_absent and not publish_absent:
                continue
            if first_only and not battery_state.is_absent:
                messages.append({
                    'topic': topic_prefix,
                    'payload': json.dumps(battery_state.to_dict(), default=str),
                })
                break
            messages.append({
                'topic': f'{topic_prefix}/{battery_state.battery_index}',
                'payload': json.dumps(battery_state.to_dict(), default=str),
            })
        mqtt.multiple(messages, hostname)


@cli.command()
def generate_systemd(**kwargs):
    """Generates a systemd unit file

    This command takes the same flags as the "run" command and will embed them in the systemd unit."""
    ctx = click.get_current_context()
    provided_options = {k: v for k, v in kwargs.items() if ctx.get_parameter_source(k) == ParameterSource.COMMANDLINE}

    click.echo(dedent(f"""\
    [Unit]
    Description=Pylon Battery to MQTT publisher
    After=network.target
    
    [Service]
    ExecStart={sys.argv[0]} run {' '.join(f'--{k}' if v is True else f'--{k}={v}' for k, v in provided_options.items())}
    Restart=on-failure
    
    [Install]
    WantedBy=multi-user.target"""))


# Give generate_systemd the exact same parameters as run
generate_systemd.params = run.params


@cli.command()
@click.option('--serial', '-s', required=True, help='Serial port (e.g., COM1, /dev/ttyUSB0) or USB device serial number')
def reset(serial: str):
    """Reset the serial device

    This is useful for cases where you have a cheap clone USB->Serial adapter that sometimes drops data. Resetting
    it usually fixes the issue for a while. Currently only works on Linux."""
    serial_port = resolve_serial_port(serial)
    reset_device(serial_port)


@cli.command()
@click.option('--serial', '-s', required=True, help='Serial port (e.g., COM1, /dev/ttyUSB0) or USB device serial number')
@click.option('--symlink_name', '-l', help='Serial port', default='pylonBattery')
def generate_udev_rule(serial: str, symlink_name: str):
    """Generates a udev rule to use a stable symlink for your serial adapter

    You can for example pipe the output of this command to /etc/udev/rules.d/99-pylonBattery.custom.rules
    to get a udev rule that will make the device appear at a predictable /dev path. This way, you
    don't have to worry about having multiple USB->Serial adapters plugged in as the one connected
    to your battery will get the same name based on its serial number.
    """
    serial_port = resolve_serial_port(serial)
    sn = get_device_serial(serial_port)
    click.echo(f'ACTION=="add", SUBSYSTEM=="tty", ATTRS{{serial}}=="{sn}", SYMLINK+="{symlink_name}"')


def main():
    logging.basicConfig(level=logging.INFO)
    try:
        cli()
    except Exception as e:
        click.echo(click.style('An unexpected error occurred during execution:', fg='red'), err=True)
        for line in traceback.format_exception(e):
            click.echo('  ' + line, err=True)
        return 1



import time
from typing import Any

import serial
from collections.abc import Iterator
from datetime import datetime
from .usbutils import reset_device

import logging
logger = logging.getLogger(__name__)


PWR_FIELDS = {
    b'Power': {
        'property': 'battery_index',
        'converter': int
    },
    b'Volt': {
        'property': 'voltage',
        'converter': lambda x: float(x) / 1000.0
    },
    b'Curr': {
        'property': 'current',
        'converter': lambda x: float(x) / 1000.0
    },
    b'Tempr': {
        'property': 'temperature',
        'converter': lambda x: float(x) / 1000.0
    },
    b'Tlow': {
        'property': 'coldest_cell_temperature',
        'converter': lambda x: float(x) / 1000.0
    },
    b'Tlow.Id': {
        'property': 'coldest_cell_id',
        'converter': lambda x: int(x)
    },
    b'Thigh': {
        'property': 'hottest_cell_temperature',
        'converter': lambda x: float(x) / 1000.0
    },
    b'Thigh.Id': {
        'property': 'hottest_cell_id',
        'converter': lambda x: int(x)
    },
    b'Vlow': {
        'property': 'lowest_cell_voltage',
        'converter': lambda x: float(x) / 1000.0
    },
    b'Vlow.Id': {
        'property': 'lowest_cell_id',
        'converter': lambda x: int(x)
    },
    b'Vhigh': {
        'property': 'highest_cell_voltage',
        'converter': lambda x: float(x) / 1000.0
    },
    b'Vhigh.Id': {
        'property': 'highest_cell_id',
        'converter': lambda x: int(x)
    },
    b'Base.St': {
        'property': 'base_status',
        'converter': lambda x: x.decode('ascii')
    },
    b'Volt.St': {
        'property': 'voltage_status',
        'converter': lambda x: x.decode('ascii')
    },
    b'Curr.St': {
        'property': 'current_status',
        'converter': lambda x: x.decode('ascii')
    },
    b'Temp.St': {
        'property': 'temperature_status',
        'converter': lambda x: x.decode('ascii')
    },
    b'Coulomb': {
        'property': 'state_of_charge',
        'converter': lambda x: int(x[:-1])
    },
    b'Time': {
        'property': 'time',
        'converter': lambda x: datetime.strptime(x.decode('ascii'), '%Y-%m-%d %H:%M:%S')
    },
    b'B.V.St': {
        'property': 'b_v_status',
        'converter': lambda x: x.decode('ascii')
    },
    b'B.T.St': {
        'property': 'b_t_status',
        'converter': lambda x: x.decode('ascii')
    },
    b'MosTempr': {
        'property': 'mosfet_temperature',
        'converter': lambda x: float(x) / 1000.0
    },
    b'M.T.St': {
        'property': 'm_t_status',
        'converter': lambda x: x.decode('ascii')
    },
    b'SysAlarm.St': {
        'property': 'sysalarm_status',
        'converter': lambda x: x.decode('ascii')
    },
}


class BatteryState:
    battery_index: int
    voltage: float
    current: float
    temperature: float
    coldest_cell_temperature: float
    coldest_cell_id: int
    hottest_cell_temperature: float
    hottest_cell_id: int
    lowest_cell_voltage: float
    lowest_cell_id: int
    highest_cell_voltage: float
    highest_cell_id: int
    base_status: str
    voltage_status: str
    current_status: str
    temperature_status: str
    state_of_charge: int
    time: datetime
    b_v_status: str
    b_t_status: str
    mosfet_temperature: float
    m_t_status: str
    sysalarm_status: str

    @classmethod
    def from_pwr_output(cls, header: list[bytes], fields: list[bytes]) -> 'BatteryState':
        res = BatteryState()

        # The output of absent batteries makes no sense - it does have some empty field
        # markers but they neither align consistently nor are the right count. So let's just
        # look for "Absent" to detect this and fill everything with None.
        if b'Absent' in fields:
            for name in header:
                setattr(res, PWR_FIELDS[name]['property'], None)
            res.battery_index = int(fields[0])
            return res

        # Since there doesn't seem to be any proper escaping, the Time ends up in two fields
        if len(header) + 1 == len(fields) and header[17] == b'Time':
            fields[17] += b' ' + fields[18]
            del fields[18]

        # After that fix, the number of fields and headers should match
        if len(header) != len(fields):
            raise ValueError('Expected %d fields, got %d' % (len(header), len(fields)))

        # Parse the expected fields, with some tolerance for None values
        for name, value in zip(header, fields):
            if name not in PWR_FIELDS:
                logger.warning('Unrecognized field %r', name)
                continue
            if value == b'-':
                setattr(res, PWR_FIELDS[name]['property'], None)
            elif value == b'Absent':
                setattr(res, PWR_FIELDS[name]['property'], None)
            else:
                setattr(res, PWR_FIELDS[name]['property'], PWR_FIELDS[name]['converter'](value))
        return res

    def to_dict(self) -> dict[str, Any]:
        return {k: getattr(self, k) for k in [f['property'] for f in PWR_FIELDS.values()]}

    @property
    def is_absent(self) -> bool:
        return self.voltage is None


class Protocol:
    def __init__(self, serial_port: str | serial.Serial):
        if isinstance(serial_port, str):
            self.serial = serial.Serial(port=serial_port, baudrate=115200, timeout=2)
        else:
            self.serial = serial_port
        logger.info('Starting protocol on %s', self.serial.port)

    def read_updates(self, interval_secs: float = 5) -> Iterator[list[BatteryState]]:
        while True:
            yield self.read_update()
            time.sleep(interval_secs)

    def read_update(self) -> list[BatteryState]:
        self.get_prompt()
        self.serial.write(b'pwr\n')
        pwr = self.serial.read_until(b'pylon>')
        lines = pwr.split(b'\r\n')
        if len(lines) != 21:
            logger.error('Expected 21 lines of pwr output, got %d: %r', len(lines), pwr)
            raise Exception('Received malformed data')

        if lines[-1] != b'\rpylon>':
            logger.error('Expected pylon> prompt after response, got %r', lines[-1])
            raise Exception('Received malformed data')

        if lines[-2] != b'\r$$':
            logger.error('Expected response to end with $$ line, got %r', lines[-2])
            raise Exception('Received malformed data')

        if lines[-3] != b'\rCommand completed successfully':
            logger.error('Command did not complete successfully, got %r', lines[-3])
            raise Exception('Received malformed data')

        header = lines[1].split()
        return [BatteryState.from_pwr_output(header, line.split()) for line in lines[2:18]]

    def have_prompt(self) -> bool:
        """Check that we have a pylon> prompt"""
        self.serial.write(b'\n')
        prompt = self.serial.read_until(b'>')
        return prompt.endswith(b'pylon>')

    def get_prompt(self):
        """Make sure we have a pylon> prompt. Try to get one if we don't."""
        if self.have_prompt():
            return True

        # Is the adapter wedged? Try resetting it
        logger.warning('No pylon> prompt, trying to reset the serial port')
        self.reset_port()
        if self.have_prompt():
            return True

        # Does the console need to be unlocked?
        logger.warning('Still no pylon> prompt, trying to unlock it with the magic string')
        self.serial.baudrate = 1200
        self.serial.write(b'~20014682C0048520FCC3\r')
        time.sleep(1)
        self.serial.baudrate = 115200
        return self.have_prompt()

    def reset_port(self):
        """Try to reset the serial port."""
        self.serial.close()
        reset_device(self.serial.port)
        time.sleep(0.1)
        self.serial.open()



from serial.tools import list_ports

# For resetting USB devices, we need udev on Linux
try:
    import pyudev
    import fcntl
except ImportError:
    pyudev = None
    fcntl = None

import logging
logger = logging.getLogger(__name__)


def reset_device(path: str) -> bool:
    """Reset the USB device.

    Returns: True if successful, False otherwise."""
    if pyudev and fcntl:
        return reset_linux_device(path)
    else:
        logger.info('Unable to reset USB device since no known methods are supported')
        return False


def reset_linux_device(path: str) -> bool:
    try:
        context = pyudev.Context()
        device = pyudev.Devices.from_device_file(context, path)
    except pyudev.DeviceNotFoundByFileError:
        return False

    # This should be a tty device, but we need to find the USB device parent
    while (device.subsystem != 'usb' or 'DEVNAME' not in device) and device.parent:
        device = device.parent
    if device.subsystem != 'usb' and 'DEVNAME' not in device:
        logger.error('Didn\'t find USB parent device, went all the way to: ' + str(device))
        return False

    logger.info('Resetting ' + str(device['DEVNAME']))
    with open(device['DEVNAME'], 'w') as f:
        try:
            r = fcntl.ioctl(f, 21780)
            if r != 0:
                logger.error('Unable to reset USB device, ioctl returned ' + str(r))
                return False
        except OSError as e:
            logger.error('Unable to reset USB device, ioctl failed: ' + str(e))

    return True


def get_device_serial(path: str) -> str:
    """Get the serial number of the USB device."""
    for port in list_ports.comports():
        if port.device == path and port.serial_number:
            return port.serial_number
    raise Exception('Unable to find serial number of device')


def find_device_by_serial(serial_number: str) -> str:
    """Find a serial device path by USB serial number."""
    for port in list_ports.comports():
        if port.serial_number == serial_number:
            return port.device
    raise Exception(f'Unable to find device with serial number: {serial_number}')



from unittest.mock import Mock
from datetime import datetime

from pylon2mqtt.protocol import Protocol

PWR_RESPONSE = (
    b'pwr\n\r@\r\r\n'
    b'Power Volt   Curr   Tempr  Tlow   Tlow.Id  Thigh  Thigh.Id Vlow   Vlow.Id  Vhigh  Vhigh.Id Base.St  Volt.St  '
    b'Curr.St  Temp.St  Coulomb  Time                 B.V.St   B.T.St  MosTempr M.T.St   SysAlarm.St\r\r\n'
    
    b'1     49874  -642   19100  17000  8        17600  0        3324   2        3325   0        Dischg   Normal   '
    b'Normal   Normal   81%      2025-03-18 05:58:51  Normal   Normal  18200    Normal   Normal  \r\r\n'
    b'2     -      -      -      -      -      -      -      Absent   -        -        -        -        '
    b'-                    -        -       \r\r\n'
    b'3     -      -      -      -      -      -      -      Absent   -        -        -        -        '
    b'-                    -        -       \r\r\n'
    b'4     -      -      -      -      -      -      -      Absent   -        -        -        -        '
    b'-                    -        -       \r\r\n'
    b'5     -      -      -      -      -      -      -      Absent   -        -        -        -        '
    b'-                    -        -       \r\r\n'
    b'6     -      -      -      -      -      -      -      Absent   -        -        -        -        '
    b'-                    -        -       \r\r\n'
    b'7     -      -      -      -      -      -      -      Absent   -        -        -        -        '
    b'-                    -        -       \r\r\n'
    b'8     -      -      -      -      -      -      -      Absent   -        -        -        -        '
    b'-                    -        -       \r\r\n'
    b'9     -      -      -      -      -      -      -      Absent   -        -        -        -        '
    b'-                    -        -       \r\r\n'
    b'10    -      -      -      -      -      -      -      Absent   -        -        -        -        '
    b'-                    -        -       \r\r\n'
    b'11    -      -      -      -      -      -      -      Absent   -        -        -        -        '
    b'-                    -        -       \r\r\n'
    b'12    -      -      -      -      -      -      -      Absent   -        -        -        -        '
    b'-                    -        -       \r\r\n'
    b'13    -      -      -      -      -      -      -      Absent   -        -        -        -        '
    b'-                    -        -       \r\r\n'
    b'14    -      -      -      -      -      -      -      Absent   -        -        -        -        '
    b'-                    -        -       \r\r\n'
    b'15    -      -      -      -      -      -      -      Absent   -        -        -        -        '
    b'-                    -        -       \r\r\n'
    b'16    -      -      -      -      -      -      -      Absent   -        -        -        -        '
    b'-                    -        -       \r\n'
    b'\rCommand completed successfully\r\n'
    b'\r$$\r\n'
    b'\rpylon>')


def test_parsing():
    serial = Mock()
    serial.write = Mock(side_effect=lambda x: len(x))
    serial.read_until = Mock(side_effect=[b'pylon>', PWR_RESPONSE])
    protocol = Protocol(serial)
    result = protocol.read_update()
    assert result[0].voltage == 49.874
    assert result[0].current == -0.642
    assert result[1].voltage is None


def test_dict_conversion():
    serial = Mock()
    serial.write = Mock(side_effect=lambda x: len(x))
    serial.read_until = Mock(side_effect=[b'pylon>', PWR_RESPONSE])
    protocol = Protocol(serial)
    result = protocol.read_update()
    assert result[0].to_dict() == {
        'battery_index': 1,
        'voltage': 49.874,
        'current': -0.642,
        'temperature': 19.1,
        'coldest_cell_temperature': 17.0,
        'coldest_cell_id': 8,
        'hottest_cell_temperature': 17.6,
        'hottest_cell_id': 0,
        'lowest_cell_voltage': 3.324,
        'lowest_cell_id': 2,
        'highest_cell_voltage': 3.325,
        'highest_cell_id': 0,
        'base_status': 'Dischg',
        'voltage_status': 'Normal',
        'current_status': 'Normal',
        'temperature_status': 'Normal',
        'state_of_charge': 81,
        'time': datetime(2025, 3, 18, 5, 58, 51),
        'b_v_status': 'Normal',
        'b_t_status': 'Normal',
        'mosfet_temperature': 18.2,
        'm_t_status': 'Normal',
        'sysalarm_status': 'Normal',
    }


def test_dict_conversion_absent_battery():
    serial = Mock()
    serial.write = Mock(side_effect=lambda x: len(x))
    serial.read_until = Mock(side_effect=[b'pylon>', PWR_RESPONSE])
    protocol = Protocol(serial)
    result = protocol.read_update()
    assert result[1].to_dict() == {
        'battery_index': 2,
        'voltage': None,
        'current': None,
        'temperature': None,
        'coldest_cell_temperature': None,
        'coldest_cell_id': None,
        'hottest_cell_temperature': None,
        'hottest_cell_id': None,
        'lowest_cell_voltage': None,
        'lowest_cell_id': None,
        'highest_cell_voltage': None,
        'highest_cell_id': None,
        'base_status': None,
        'voltage_status': None,
        'current_status': None,
        'temperature_status': None,
        'state_of_charge': None,
        'time': None,
        'b_v_status': None,
        'b_t_status': None,
        'mosfet_temperature': None,
        'm_t_status': None,
        'sysalarm_status': None,
    }
